/* tdmflm - fluid mechanics for tdm */

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "tdmdef.h"
#include "cdata.h"
#include "cimped.h"
#include "cmatrx.h"
#include "cmider.h"
#include "cparam.h"

double xinterp(double, float *);

static float r1, r2, r4, astom;

/* mxprn - print matrix (for debugging) */

static void
mxprn(int m, float *x)
{
    int i, j;

    for (i = 0; i < m; i++) {
        for (j = 0; j < m; j++) {
            printf("%8.3f", x[i + j * m]);
	}
	printf("\n");
    }
    printf("\n");
}

/* mxsol1d - partially invert matrix for 1D fluid */

static void
mxsol1d()
{
    float   xol;
    int     i, k, nn = pa_.n - 1;

    alfx = astom;

    /* compute fluid mass */ 
    for (i = 0; i <= nn; i++) {
	xol = i / (float) nn;
        z0m[i] = (2 * pa_.rho * dx) / xinterp(xol, &pa_.aca);
	abmom[i] = xinterp(xol, &pa_.bwa) * dx / z1m[i]; 
    } 
    if (pa_.apxacc) {
	abmom[nn] = 1 / pa_.mhe;
    }

    /* initialize matrix */
    aa1[0] = 0;
    if (pa_.basacc) {
        aa2[0] = 1 / z0m[0] + abmom[0] + alfx;
        aa3[0] = -1 / z0m[0];
    } else {
        aa2[0] = 1;
        aa3[0] = 0;
    }
    if (nn <=0)
    	return;
    for (k = 1; k < nn; k++) {
	aa1[k] = -1 / z0m[k - 1];
	aa2[k] =  1 / z0m[k - 1] + 1 / z0m[k] + abmom[k];
	aa3[k] = -1 / z0m[k];
    }
    if (pa_.apxacc) {
        aa1[nn] = -1 / z0m[nn - 1];
        aa2[nn] =  1 / z0m[nn - 1] + abmom[nn];
    } else {
        aa1[nn] = 0;
        aa2[nn] = 1;
    }
    aa3[nn] = 0; 

    /* partial inversion */
    aa2[0] = 1 / aa2[0];
    aa3[0] *= aa2[0];
    for (k = 1; k < nn; k++) {
	aa2[k] = 1 / (aa2[k] - aa1[k] * aa3[k - 1]);
	aa3[k] *= aa2[k];
    }
    if (pa_.apxacc) {
	aa2[nn] = 1 / (aa2[nn] - aa1[nn] * aa3[nn - 1]);
	aa3[nn] *= aa2[nn];
    }
}

/* mxadd - add next submatrix. */

static void
mxadd(float *ak, int k)
{
    int i, m, msq, mm1;

    m = pa_.m;
    msq = m * m;
    mm1 = m - 1;

    /*  add kth sub-matrix to whatever was copied into kth place. */
    ak[0] += r4 + abmom[k];
    ak[m] -= r2;
    for (i = 1; i < mm1; i++) {
        ak[(i - 1) * m + i] -= r1;
        ak[(i    ) * m + i] += r4;
        ak[(i + 1) * m + i] -= r1;
    }
    ak[(mm1 - 1) * m + mm1] -= r2;
    ak[(mm1    ) * m + mm1] += r4;

}

/* mxinv - invert the kth sub-matrix ak(i,j) */

static void
mxinv(int m, float *ak)
{
    double am;
    float b[Bsize], at;
    int i, j, ix1, jx1, ix2, jx2, ji, mm1, mp1, mx1, msq;

    mm1 = m - 1;
    mp1 = m + 1;
    mx1 = mm1 * m;
    msq = m * m;

    /*  work down from top */
    for (j = 0; j < m; j++) {
        jx1 = j * m;
        ix1 = j + jx1;
        at = ak[ix1];
        am = fabs(at);
        if (am < 1e-9) {
	    fprintf(stderr, "*** Ill-conditioned matrix! (%.3g)\n", am);
	    mxprn(m, ak);
	    exit(1);
	}
        ak[ix1] = 1;
        for (ix2 = j; ix2 < msq; ix2 += m)
            ak[ix2] /= at;
        if (j < mm1) {
            for (i = (j + 1); i < m; i++) {
                at = -ak[i + jx1];
                ak[i + jx1] = 0;
                for (jx2 = 0; jx2 <= mx1; jx2 += m) {
                    ak[i + jx2] += ak[j + jx2] * at;
                }
            }
        }
    }

    /* copy lower triangle into 'b' */
    for (i = 0; i < msq; i++)
        b[i] = ak[i];

    /* zero upper triangle */
    for (i = 0; i < mm1; i++)
        for (ix2 = (i * mp1 + m); ix2 < msq; ix2 += m)
            b[ix2] = 0;

    /* work up from bottom */
    for (i = m - 2; i >= 0; i--) {
        for (ji = i + 1; ji < m; ji++) {
            jx1 = ji * m;
            for (j = 0; j < m; j++) {
                jx2 = j * m;
                b[i + jx2] -= ak[i + jx1] * b[ji + jx2];
            }
        }
    }

    /* put inverted matrix back into place */
    for (i = 0; i < msq; i++)
        ak[i] = b[i];
}

/* mxbas - handles sub-matrices at base */

static void
mxbas()
{
    float qst, *ak, *bk;
    int i, j, k, m, mm1, mp1, msq;
   
    m = pa_.m;
    k = 1;
    mm1 = m - 1;
    mp1 = m + 1;
    msq = m * m;
    qst = alfx / mm1;

    /* do first submatrix */
    ak = aaa;
    if (pa_.basacc) {			/*  acceleration boundary */
        for (i = 0;  i < m; i++) {	/*  middle-ear influence */
            ak[i] = qst / 2;
            for (j = 1;  j < mm1; j++)
                ak[i + j * m] = qst;
            ak[i + mm1 * m] = qst / 2;
	}
       mxadd(ak, 0);
       mxinv(m, ak);
    } else { 				/* pressure boundary */
        for (i = 0;  i < msq; i++)
            ak[i] = 0;
        for (i = 0;  i < m; i++)
            ak[i * mp1] = 1;
    }

   /*  do second sub-matrix */
   bk = aaa + msq;
   if(pa_.basacc) 		/* acceleration boundary */
	for (i = 0; i < msq; i++)
            bk[i] = -2 * ak[i];
   else       			/* pressure boundary */
	for (i = 0; i < msq; i++)
            bk[i] = 0;
   mxadd(bk, 1);
   mxinv(m, bk);

 }

/* mxmid - handles sub-matrices between apex and base */

static void
mxmid()
{
    float *ak, *bk;
    int i, k, m, msq, nn;

    m = pa_.m;
    nn = pa_.n - 1;
    msq = m * m;

    for (k = 2; k < nn; k++) {
	ak = aaa + msq * (k - 1);
	bk = aaa + msq * k;
        /* copy sub-matrix (k-1) into sub-matrix k. */
        for (i = 0; i < msq; i++)
           bk[i] = -ak[i];
        mxadd(bk, k);
        mxinv(m, bk);
   }

}

/* mxapx - handles sub-matrix at apex */

static void
mxapx()
{
    float *ak, *bk;
    int i, m, mp1, msq, nn;

    m = pa_.m;
    nn = pa_.n - 1;
    mp1 = m + 1;
    msq = m * m;

    ak = aaa + msq * (nn - 1);
    bk = aaa + msq * nn;
    if (pa_.apxacc) {			/*  acceleration boundary */
        /* copy  sub-matrix (nn-1) times 2 into sub-matrix nn. */
        for (i = 0; i < msq; i++)
           bk[i] = -2 * ak[i];
        mxadd(bk, nn);
        mxinv(m, bk);
    } else {
        for (i = 0;  i < msq; i++)
            bk[i] = 0;
        for (i = 0;  i < m; i++)
            bk[i * mp1] = 1;
    }

}

/* mxsol2d - partially invert matrix for 2D fluid */

static void
mxsol2d()
{
    float   xol, wr;
    int     i, nn = pa_.n - 1;

    alfx = ((4 * pa_.rho * dx) / pa_.aca) * astom;
    r1 = (dx * dx) / (dz * dz);
    r2 = r1 + r1;
    r4 = 2. + r2;

    /* compute fluid/bm mass ratio */ 
    for (i = 0; i <= nn; i++) {
	xol = i / (float) nn;
	wr = xinterp(xol, &pa_.bwa) / pa_.yw; 
        abmom[i] = (4 * pa_.rho * dz) * r1 * wr / z1m[i]; 
    }

    mxbas();	/*  start at basal boundary */
    mxmid();	/*  base to apex */
    mxapx();	/*  do apical boundary */
}

/* mechfl - initialize fluid mechanics */

void
mechfl()
{
    aapx = pa_.apxacc ? 1 : 0;
    abas = pa_.basacc ? 1 : 0;
    papx = pa_.apxacc ? 0 : 1;
    pbas = pa_.basacc ? 0 : 1;
    astom = ame[ME_ST] * smem[ME_ST][ME_ST];

    if (pa_.n <= 1) {
	z1m[0] = pa_.mst / pa_.ast;
	z2m[0] = pa_.mco * pa_.ast;
    } else {
        if (pa_.m == 1) {
	    mxsol1d();
	} else {
	    mxsol2d();
        }
    }
}

